#define _CRT_SECURE_NO_DEPRECATE

#include <windows.h>

#include "filewriter_sync.h"


unsigned char* FileWriterSync::mBuffer = NULL;


FileWriterSync* FileWriterSync::GetFileWriter(const std::string& inName, long long inFileSize)
{
	if (!mBuffer)
		mBuffer = new unsigned char[mMaxMemSize];
	return new FileWriterSync(inName, inFileSize);
}


void FileWriterSync::Finalize()
{
	delete mBuffer;
}


FileWriterSync::FileWriterSync(const std::string& inName, long long inFileSize) :
	FileWriter(inName, inFileSize),
	mHandle(INVALID_HANDLE_VALUE),
	mLength(0)
{
}


bool FileWriterSync::Open(bool inWriteable, int inModTime)
{
	mLength = 0;
	mHandle = CreateFile(inWriteable, false, inModTime);
	return mHandle != INVALID_HANDLE_VALUE;
}


bool FileWriterSync::Close()
{
	bool flush_ok = Flush();
	bool close_res = CloseFile(mHandle);
	return flush_ok && close_res;
}


void FileWriterSync::GetWriteBuffer(unsigned char*& outBuf, int& outLength)
{
	outBuf = mBuffer + mLength;
	outLength = mMaxMemSize - mLength;
}


bool FileWriterSync::ReleaseWriteBuffer(int inRemainingLength)
{
	mLength = mMaxMemSize - inRemainingLength;
	// We need to have at least two bytes free, as we may need to do a line-end conversion
	if (mLength >= mMaxMemSize - 1)
		return Flush();
	return true;
}


bool FileWriterSync::Flush()
{
	Log("Flush\n");
	if (mLength > 0)
	{
		DWORD written;
		if (::WriteFile(mHandle, mBuffer, mLength, &written, NULL) == 0 || written != mLength)
			return false;
		if (mWriteCallback)
			mWriteCallback(written);
		mLength = 0;
	}
	return true;
}


void FileWriterSync::HandleAbort()
{
	CloseFile(mHandle);
	// Aborted before file actions could be finished, so delete
	DeleteFile();
}


bool FileWriterSync::ChModTime(int inModTime)
{
	return FileWriter::ChModTime(mHandle, inModTime);
}
