#define _CRT_SECURE_NO_DEPRECATE
#include "windows.h"
#include <stdio.h>
#include "fs.h"


void sync_stat_callback(int inFileCount, long long inSyncSize)
{
	Log("Syncing %d files, with a total size of %d bytes.\n", inFileCount, inSyncSize);
}


void file_stat_callback(const char* inDepotFile, const char* inClientFile, const char* inAction, int inRev, long long inFileSize)
{
	const char* action = inAction;
	// Do some translation, so we produce the exact same output p4 would
	if (strcmp(action, "updated") == 0)
		action = "updating";
	if (strcmp(action, "refreshed") == 0)
		action = "refreshing";
	else if (strcmp(action, "added") == 0)
		action = "added as";
	else if (strcmp(action, "deleted") == 0)
		action = "deleted as";
	if (DoScriptedOutput())
		printf("info: ");
	Print("%s#%d - %s %s\n", inDepotFile, inRev, action, inClientFile);
}


BOOL WINAPI ExceptionHandler(DWORD dwCtrlType)
{
	AbortSync();
	ExitProcess(FS_ABORTED);
	return TRUE;
}


int main(int argc, char **argv)
{
	SetConsoleCtrlHandler(ExceptionHandler, TRUE);
	EResult result = Sync(argc, argv, "p4fs", sync_stat_callback, file_stat_callback);
	SetConsoleCtrlHandler(ExceptionHandler, FALSE);

	if (result == FS_HELP)
	{
		Print(GetErrorMessage());
	}
	else if (result == FS_ERROR_NO_SYNC)
	{
		// Unsupported p4 command, redirect to p4 itself
		if (!DoScriptedOutput())
			Print("Command unsupported by p4fs, redirecting to p4.\n");
		STARTUPINFO start_up_info;
		::memset(&start_up_info, 0, sizeof(STARTUPINFO));
		start_up_info.cb = sizeof(STARTUPINFO);
		PROCESS_INFORMATION process_information;
		std::string command_line("p4");
		for (int i = 1; i < argc; ++i)
		{
			command_line += " ";
			command_line += argv[i];
		}
		if (command_line.length() > 32768)
		{
			if (DoScriptedOutput())
				printf("error: ");
			Print("Command line too long, maximum length = 32768 characters.\n");
			return FS_ERROR_ARGUMENT;
		}
		BOOL start_p4 = ::CreateProcess(NULL, (char*) command_line.c_str(), NULL, NULL, TRUE, 0, NULL, NULL, &start_up_info, &process_information);
		if (start_p4)
		{
			::WaitForSingleObject(process_information.hProcess, INFINITE);
			DWORD exit_code;
			::GetExitCodeProcess(process_information.hProcess, &exit_code);
			::CloseHandle(process_information.hProcess);
			::CloseHandle(process_information.hThread);
			return exit_code;
		}
		else
		{
			if (DoScriptedOutput())
				printf("error: ");
			char* message = NULL;
			DWORD last_err = ::GetLastError();
			if (::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, last_err, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR) &message, 0, NULL) > 0)
			{
				Print("Error starting p4 : %s\n", message);
				LocalFree(message);
			}
			else
				Print("Error starting p4, error code = %d.\n", last_err);
			result = FS_ERROR_P4; 
		}
	}
	else if (result != FS_OK)
	{
		if (DoScriptedOutput())
			fprintf(stderr, "error: ");
		fprintf(stderr, GetErrorMessage());
		fflush(stderr);
	}

	if (result == FS_HELP)
		result = FS_OK;

	if (DoScriptedOutput())
		printf("exit: %d\n", result);
		
	return result;
}
